#!/bin/sh

# Configuration shell script

TARGET=mpsl

# gets program version
VERSION=`cut -f2 -d\" VERSION`

# default installation prefix
PREFIX=/usr/local

# installation directory for documents
DOCDIR=""

# store command line args for configuring the libraries
CONF_ARGS="$*"

MINGW32_PREFIX=i686-w64-mingw32

C_MPSL_OBJS="mpsl_l.o mpsl_y.o"

CONFOPT_DEFAULT_COMPILER="a_mpsl"

# parse arguments
while [ $# -gt 0 ] ; do

    case $1 in
    --help)         CONFIG_HELP=1 ;;

    --mingw32-prefix=*)     MINGW32_PREFIX=`echo $1 | sed -e 's/--mingw32-prefix=//'`
                            ;;

    --mingw32)      CC=${MINGW32_PREFIX}-gcc
                    WINDRES=${MINGW32_PREFIX}-windres
                    AR=${MINGW32_PREFIX}-ar
                    LD=${MINGW32_PREFIX}-ld
                    CPP=${MINGW32_PREFIX}-g++
                    CFLAGS="-O3 $CFLAGS"
                    ;;

    --prefix)       PREFIX=$2 ; shift ;;
    --prefix=*)     PREFIX=`echo $1 | sed -e 's/--prefix=//'` ;;

    --docdir)       DOCDIR=$2 ; shift ;;
    --docdir=*)     DOCDIR=`echo $1 | sed -e 's/--docdir=//'` ;;

    --compiler)     CONFOPT_DEFAULT_COMPILER=$2 ; shift ;;
    --compiler=*)   CONFOPT_DEFAULT_COMPILER=`echo $1 | sed -e 's/--compiler=//'` ;;

    esac

    shift
done

if [ "$CONFIG_HELP" = "1" ] ; then

    echo "Available options:"
    echo "--prefix=PREFIX         Installation prefix ($PREFIX)."
    echo "--docdir=DOCDIR         Instalation directory for documentation."
    echo "--compiler=COMPILER     Set default compiler. Values can be:"
    echo "                        c_mpsl (classic compiler, needs flex and yacc/bison)"
    echo "                        a_mpsl (ad-hoc compiler)"
    echo "--mingw32-prefix=PREFIX Prefix name for mingw32 ($MINGW32_PREFIX)."
    echo "--mingw32               Build using the mingw32 compiler."

    echo
    echo "Environment variables:"
    echo "CC                    C Compiler."
    echo "AR                    Library Archiver."
    echo "YACC                  Parser."

    exit 1
fi

if [ "$DOCDIR" = "" ] ; then
    DOCDIR=$PREFIX/share/doc/mpsl
fi

echo "Configuring MPSL..."

echo "/* automatically created by config.sh - do not modify */" > config.h
echo "# automatically created by config.sh - do not modify" > makefile.opts
> config.ldflags
> config.cflags
> .config.log

# set compiler
if [ "$CC" = "" ] ; then
    CC=cc
    # if CC is unset, try if gcc is available
    which gcc > /dev/null 2>&1

    if [ $? = 0 ] ; then
        CC=gcc
    fi
fi

echo "CC=$CC" >> makefile.opts

# set archiver
if [ "$AR" = "" ] ; then
    AR=ar
fi

echo "AR=$AR" >> makefile.opts

# set parser
if [ "$YACC" = "" ] ; then
    YACC=yacc
fi

echo "YACC=$YACC" >> makefile.opts

# add version
cat VERSION >> config.h

# add installation prefix
echo "#define CONFOPT_PREFIX \"$PREFIX\"" >> config.h

#########################################################

# configuration directives

# CFLAGS
if [ -z "$CFLAGS" ] ; then
    CFLAGS="-g -Wall"
fi

echo -n "Testing if C compiler supports ${CFLAGS}... "
echo "int main(int argc, char *argv[]) { return 0; }" > .tmp.c

$CC .tmp.c -o .tmp.o 2>> .config.log

if [ $? = 0 ] ; then
    echo "OK"
else
    echo "No; resetting to defaults"
    CFLAGS=""
fi

echo "CFLAGS=$CFLAGS" >> makefile.opts

# Add CFLAGS to CC
CC="$CC $CFLAGS"

# MPDM
echo -n "Looking for MPDM... "

for MPDM in ./mpdm ../mpdm NOTFOUND ; do
    if [ -d $MPDM ] && [ -f $MPDM/mpdm.h ] ; then
        break
    fi
done

if [ "$MPDM" != "NOTFOUND" ] ; then
    echo "-I$MPDM" >> config.cflags
    echo "-L$MPDM -lmpdm" >> config.ldflags
    echo "OK ($MPDM)"
else
    echo "No"
    exit 1
fi

echo
(cd $MPDM ; ./config.sh $CONF_ARGS)
echo

# test if flex and yacc are available
echo -n "Testing for flex and $YACC... "
if ! (which flex && which $YACC) > /dev/null 2>&1 ; then
    echo "No; classic compiler not available"
    echo "#define CONFOPT_WITHOUT_C_MPSL 1" >> config.h
    C_MPSL_OBJS=""
    CONFOPT_DEFAULT_COMPILER="a_mpsl"
else
    echo "Found; classic compiler is available"
fi

if [ ! -z "$CONFOPT_DEFAULT_COMPILER" ] ; then
    echo "#define CONFOPT_DEFAULT_COMPILER \"$CONFOPT_DEFAULT_COMPILER\"" >> config.h
fi

cat $MPDM/config.ldflags >> config.ldflags
echo "-lm" >> config.ldflags

# if win32, the interpreter is called mpsl.exe
grep CONFOPT_WIN32 ${MPDM}/config.h >/dev/null && TARGET=${TARGET}.exe


#########################################################

# final setup

echo "MPDM=$MPDM" >> makefile.opts
grep DOCS $MPDM/makefile.opts >> makefile.opts
echo "VERSION=$VERSION" >> makefile.opts
echo "PREFIX=\$(DESTDIR)$PREFIX" >> makefile.opts
echo "DOCDIR=\$(DESTDIR)$DOCDIR" >> makefile.opts
echo "CONF_ARGS=$CONF_ARGS" >> makefile.opts
echo "TARGET=$TARGET" >> makefile.opts
echo "C_MPSL_OBJS=$C_MPSL_OBJS" >> makefile.opts
echo >> makefile.opts

cat makefile.opts makefile.in makefile.depend > Makefile

#########################################################

# cleanup

rm -f .tmp.c .tmp.o

exit 0
